--[[
--
--	TargetCheck
--	by Dust of Turalyon
--
--]]

TC_Version = GetAddOnMetadata("TargetCheck", "Version");
TC_SavedVersion = GetAddOnMetadata("TargetCheck", "X-SavedVersion");

TC_ClassList = {
	"MAGE";
	"WARLOCK";
	"ROGUE";
	"WARRIOR";
	"HUNTER";
	"PRIEST";
	"DRUID";
	"PALADIN";
	"SHAMAN";
	"DEATHKNIGHT";
	"PET";
};
-- Add classes as keys for fast lookup
for _,v in ipairs(TC_ClassList) do
	TC_ClassList[v] = true;
end

TC_OnLoad = function (self)
	--Slash command
	SlashCmdList["TARGETCHECK"] = TC_SlashHandler;
	SLASH_TARGETCHECK1 = "/targetcheck";
	SLASH_TARGETCHECK2 = "/tc";

	TargetCheck_MainFrame = CreateFrame("Frame", nil, UIParent);
	TargetCheck_MainFrame:SetScript("OnEvent", TC_OnEvent);

	--Events
	TargetCheck_MainFrame:RegisterEvent("ADDON_LOADED");
	TC_OnLoad = nil; -- memory usage reduction
end;

TC_OnVariablesLoaded = function ()
	if (TC_Settings == nil) or (TC_Settings["Version"] ~= TC_SavedVersion) then
		TC_MSG("TC: TargetCheck settings reset");
		local mouse = TC_CreateFrameTable();
		mouse["Anchor"] = "CURSOR";
		mouse["Target"] = "mouseover";
		TC_Settings =
		{
			["Scale"]	= 0.75;
			["Alpha"]	= 0.75;

			["Frames"]	= {};

			["MouseFrame"]	= mouse;
			["Mouse"]	= false;
			["Version"]	= TC_SavedVersion;
		}
		TC_CreateFrame();
	end;

	TC_UnitListsInvalid = true;
	TC_OnVariablesLoaded = nil; -- memory usage reduction
end;

TC_RegisterEvents = function ()
	--OnUpdate
	TargetCheck_MainFrame:SetScript("OnUpdate", TC_OnUpdate);

	--Handlers
	local invalidateTL = function () TC_TargetsInvalid = true; end;
	TC_EventHandlers["UNIT_TARGET"] = invalidateTL;
	TC_EventHandlers["PLAYER_TARGET_CHANGED"] = invalidateTL;
	TC_EventHandlers["PLAYER_FOCUS_CHANGED"] = invalidateTL;
	TC_EventHandlers["UPDATE_MOUSEOVER_UNIT"] = invalidateTL;
	local invalidateUL = function () TC_UnitListsInvalid = true; end;
	TC_EventHandlers["UNIT_PET"] = invalidateUL;
	TC_EventHandlers["RAID_ROSTER_UPDATE"] = invalidateUL;
	TC_EventHandlers["PARTY_MEMBERS_CHANGED"] = invalidateUL;
	TC_EventHandlers["PLAYER_ENTERING_WORLD"] = invalidateUL;

	--Events
	TargetCheck_MainFrame:RegisterEvent("UNIT_TARGET");
	TargetCheck_MainFrame:RegisterEvent("PLAYER_TARGET_CHANGED");
	TargetCheck_MainFrame:RegisterEvent("PLAYER_FOCUS_CHANGED");
	TargetCheck_MainFrame:RegisterEvent("UPDATE_MOUSEOVER_UNIT");
	-- TODO: Add event that fires on mouseover unit loss
	-- Currently an OnUpdate hack in TargetCheck_Tooltip is used
	--Rebuild UnitList
	TargetCheck_MainFrame:RegisterEvent("UNIT_PET");
	TargetCheck_MainFrame:RegisterEvent("RAID_ROSTER_UPDATE");
	TargetCheck_MainFrame:RegisterEvent("PARTY_MEMBERS_CHANGED");
	TargetCheck_MainFrame:RegisterEvent("PLAYER_ENTERING_WORLD");
	TC_RegisterEvents = nil; -- memory usage reduction
end;

TC_EventHandlers = {
	["ADDON_LOADED"] = function (frame, event, addonName)
			if addonName == "TargetCheck" then
				TC_OnVariablesLoaded();
				TC_RegisterEvents();
				TC_RestoreFrames();
				TC_Check();
				TargetCheck_MainFrame:UnregisterEvent("ADDON_LOADED");
				TC_EventHandlers["ADDON_LOADED"] = nil; -- Oneshot event
			end
		end;
}

TC_OnEvent = function (self, event, ...)
	local handler = TC_EventHandlers[event];
	if handler then
		handler(frame, event, ...);
	end;
end;

TC_OnUpdate = function (self, delta)
	-- If the target list is invalid, check targets
	-- Delayed check might cause one frame additional latency
	-- Causes slight idle-overhead
	-- Limits number of checks to one per frame, no matter how many events are sent
	if TC_TargetsInvalid then
		TC_Check();
	end
end;

TC_FrameList = {};

TC_CreateFrameTable = function ()
	local f = {
			["Anchor"]	= "TOPLEFT";
			["Target"]	= "target";
			["Limit"]	= 6;
			["Header"]	= true;
			["Mana"]	= false;
			["Invert"]	= false;
		};
	local sortOrder = {}
	for _, v in ipairs(TC_ClassList) do
		-- Set up default sort order as in classList
		table.insert(sortOrder, v);
		-- Enable all classes
		f[v] = true;
	end
	f["SortOrder"] = sortOrder;
	return f;
end;

TC_CreateFrame = function ()
	local f = TC_CreateFrameTable();
	table.insert(TC_Settings["Frames"], f);
	return f;
end;

TC_DeleteFrame = function (self, frame)
	local i = 1;
	while TC_FrameList[i] ~= nil and TC_FrameList[i] ~= frame do
		i = i + 1;
	end;
	table.remove(TC_Settings["Frames"], i);
	table.remove(TC_FrameList, i);
	frame["Pointer"]:Hide();
	for i = 0, 40 do
		local text = getglobal(frame["Name"].."_Text"..i);
		local texture = getglobal(frame["Name"].."_Texture"..i);
		text:Hide();
		texture:Hide();
	end;
	frame["Pointer"].deleted = true;
end;

TC_RestoreFrames = function ()
	for i, f in ipairs(TC_Settings["Frames"]) do
		TC_RestoreFrame(f, i);
	end
	TC_RestoreMouseoverFrame();
	TC_RestoreFrames = nil;
end;

TC_RestoreFrame = function (f, i)
	local name = "TC_OutputFrame"..i;
	local pointer = getglobal(name);
	while pointer do
		i = i + 1;
		name = "TC_OutputFrame"..i;
		pointer = getglobal(name);
		-- might overwrite names. not good. don't know how to fix it yet.
	end;
	pointer = CreateFrame("Frame", name, UIParent, "TargetCheck_OutputFrameTemplate");
	local frame = {
		["Settings"] = f;

		["Name"]	= name;
		["Pointer"]	= pointer;
		["Lines"]	= 0;

		["writeLine"]	= TC_WriteLine;
		["clearLines"]	= TC_ClearLines;
		["finishWrites"]= TC_FinishWrites;

		["unitList"]	= {["unitTCount"] = 0; ["unitFCount"] = 0;};
	};
	for _, class in ipairs(frame["Settings"]["SortOrder"]) do
		frame["unitList"][class] = {["unitTCount"] = 0; ["unitFCount"] = 0;};
	end;
	table.insert(TC_FrameList, frame)

	pointer:SetScale(TC_Settings["Scale"]);
	pointer:SetAlpha(TC_Settings["Alpha"]);
	pointer["TC_Frame"] = frame;
	TC_SetAnchor(frame, f["Anchor"]);
	TC_Menu_Load(_G[name.."_Menu"]);

	return frame;
end;

TC_RestoreMouseoverFrame = function ()
	TC_MouseFrame = { -- pseudo frame
		["Settings"]	= TC_Settings["MouseFrame"];
		["Name"]	= "Mouse Frame";
		["Pointer"] = TargetCheck_Tooltip;
		["Lines"]	= 0;

		["writeLine"]	= TC_WriteLineTooltip;
		["clearLines"]	= TC_ClearLinesTooltip;
		["finishWrites"]= TC_FinishWritesTooltip;

		["unitList"]	= {["unitTCount"] = 0; ["unitFCount"] = 0;};
	};
	for _, class in ipairs(TC_MouseFrame["Settings"]["SortOrder"]) do
		TC_MouseFrame["unitList"][class] = {["unitTCount"] = 0; ["unitFCount"] = 0;};
	end;
	TargetCheck_Tooltip:SetScale(TC_Settings["Scale"]);
end;

TC_UnitCacheStack = {};
TC_UnitCacheStackSize = 0;
TC_UnitCreate = function (unitID)
	if TC_UnitCacheStackSize == 0 then
		return {unitID, false};
	else
		local unit = TC_UnitCacheStack[TC_UnitCacheStackSize];
		TC_UnitCacheStack[TC_UnitCacheStackSize] = nil;
		TC_UnitCacheStackSize = TC_UnitCacheStackSize - 1;
		unit[1] = unitID;
		unit[2] = false;
		return unit;
	end;
end;
TC_UnitCache = function (unit)
	TC_UnitCacheStackSize = TC_UnitCacheStackSize + 1;
	TC_UnitCacheStack[TC_UnitCacheStackSize] = unit;
end;

TC_GenerateUnitList = function (frame)
	local unitList = frame["unitList"];
	local i = 1;
	-- Recycle tables (I know, I know...)
	for _, class in ipairs(frame["Settings"]["SortOrder"]) do
		i = 1;
		while unitList[class][i] ~= nil do
			TC_UnitCache(unitList[class][i]);
			unitList[class][i] = nil;
			i = i + 1;
		end;
	end;
	-- Rebuild table
	if GetNumRaidMembers() > 0 then
		-- For raid situations
		for i = 1, GetNumRaidMembers() do
			local unitID = "raid"..tostring(i);
			if not UnitIsUnit("player", unitID) then
				local localClass, intClass = UnitClass(unitID);
				local unitClassList = unitList[intClass];
				if unitClassList then
					table.insert(unitClassList, TC_UnitCreate(unitID));
				else
					-- Should not happen
					TC_MSG("TargetCheck warning: Got no class list for unit "..unitID..", class "..(intClass or "nil"));
				end;
				-- Pets of raiders
				local unitPetID = unitID.."pet";
				if UnitExists(unitPetID) then
					table.insert(unitList["PET"], TC_UnitCreate(unitPetID));
				end;
			end;
		end;
	else
		-- Or in parties
		for i = 1, GetNumPartyMembers() do
			local unitID = "party"..tostring(i);
			local localClass, intClass = UnitClass(unitID);
			local unitClassList = unitList[intClass];
			if unitClassList then
				table.insert(unitClassList, TC_UnitCreate(unitID));
			else
				-- Should not happen
				TC_MSG("TargetCheck warning: Got no class list for unit "..unitID..", class "..(intClass or "nil"));
			end;
			-- Pets of party members
			local unitPetID = unitID.."pet";
			if UnitExists(unitPetID) then
				table.insert(unitList["PET"], TC_UnitCreate(unitPetID));
			end;
		end;
	end;
	-- Own pet
	if UnitExists("pet") then
		table.insert(unitList["PET"], TC_UnitCreate("pet"));
	end;
	return unitList;
end;

TC_Performance = {
	["GenerateUnitList"] = 0;
	["Check"] = 0;
};

-- Only rebuild when UnitLists are invalid, checked in TC_Check
TC_UnitListsInvalid = false;
TC_RebuildUnitLists = function ()
	TC_Performance["GenerateUnitList"] = TC_Performance["GenerateUnitList"] + 1;
	for _, f in ipairs(TC_FrameList) do
		TC_GenerateUnitList(f);
	end;
	if (TC_Settings["Mouse"] and TC_MouseFrame) then
		TC_GenerateUnitList(TC_MouseFrame);
	end;
	TC_UnitListsInvalid = false;
end;

-- Only check for targets when invalid, checked in TC_OnUpdate
TC_TargetsInvalid = false;
TC_Check = function ()
	if TC_UnitListsInvalid then
		TC_RebuildUnitLists();
	end;
	TC_Performance["Check"] = TC_Performance["Check"] + 1;
	for _, f in ipairs(TC_FrameList) do
		TC_CheckTarget(f);
		TC_WriteToFrame(f);
	end;
	if (TC_Settings["Mouse"] and TC_MouseFrame) then
		TC_CheckTarget(TC_MouseFrame);
		TC_WriteToFrame(TC_MouseFrame);
	end;
	TC_TargetsInvalid = false;
end;

TC_CheckTarget = function (frame)
	local target = frame["Settings"]["Target"];
	if UnitExists(target) then
		local targetCounter = 0;
		local notTargetCounter = 0;
		-- Find the Players who target the same target you do
		for _, class in ipairs(frame["Settings"]["SortOrder"]) do
			local classUnits = frame["unitList"][class];
			classUnits["unitTCount"] = 0;
			classUnits["unitFCount"] = 0;
			if frame["Settings"][class] then
				for _, unit in ipairs(classUnits) do
					local unitID = unit[1];
					if UnitIsUnit(target, unitID.."target") then
						targetCounter = targetCounter + 1;
						classUnits["unitTCount"] = classUnits["unitTCount"] + 1;
						unit[2] = true;
					else
						notTargetCounter = notTargetCounter + 1;
						classUnits["unitFCount"] = classUnits["unitFCount"] + 1;
						unit[2] = false;
					end
				end;
			end;
		end;
		frame["unitList"]["unitTCount"] = targetCounter;
		frame["unitList"]["unitFCount"] = notTargetCounter;
	end;
end;

TC_WriteToFrame = function (frame)
	local writeCondition = not frame["Settings"]["Invert"];
	local writeCount;
	if writeCondition then
		writeCount = frame["unitList"]["unitTCount"];
	else
		writeCount = frame["unitList"]["unitFCount"];
	end;

	frame:clearLines();

	if (not UnitExists(frame["Settings"]["Target"])) or
		(writeCount == 0 and (not frame["Settings"]["Header"])) then
		frame["Pointer"]:Hide();
		return;
	else
		-- Setup Target of Target function
		if (frame["Settings"]["Header"]) then
			frame:writeLine((UnitName(frame["Settings"]["Target"])),
				1,1,1, nil, frame["Settings"]["Target"]);
--[[		if (frame["Settings"]["TargetOfTarget"]) and
			UnitExists(frame["Settings"]["Target"].."target") then
			frame:writeLine("-> "..UnitName(frame["Settings"]["Target"].."target"),
				0.75,0.75,0.75, nil, frame["Settings"]["Target"].."target");
		end;]]
		end;

		if writeCount > 0 then
			TC_Write(frame);
		end;
		frame:finishWrites();
		return;
	end
end;

TC_WriteReduced = function (frame)
	local unitList = frame["unitList"];
	local settings = frame["Settings"];
	local writeCondition = not settings["Invert"];
	local writeCountSpecifier;
	if writeCondition then
		writeCountSpecifier = "unitTCount";
	else
		writeCountSpecifier = "unitFCount";
	end;
	local classesWithPlayers = 0;	-- The number of classes that are shown
	local classesToList = 0;	-- The number of classes that are expanded
	local writeCount = unitList[writeCountSpecifier];
	-- Find out how many classes have players
	for _, class in ipairs(settings["SortOrder"]) do
		if settings[class] and (unitList[class][writeCountSpecifier] > 0) then
			classesWithPlayers = classesWithPlayers + 1;
		end;
	end;
	-- Calculate how many classes can be expanded
	local linesLeft = settings["Limit"]-frame["Lines"] - classesWithPlayers;
	for _, class in ipairs(settings["SortOrder"]) do
		local classSize = unitList[class][writeCountSpecifier];
		-- expand a class if, when expanded, there are still lines left
		if settings[class] and (classSize > 0) then
			if (linesLeft - classSize + 1 >= 0) then
				classesToList = classesToList + 1;
				linesLeft = linesLeft - classSize + 1;
			else
				break;
			end;
		end;
	end;
	-- Finaly write the list
	for _, class in ipairs(settings["SortOrder"]) do
		local classSize = unitList[class][writeCountSpecifier];
		if settings[class] and (classSize > 0) then
			linesLeft = settings["Limit"]-frame["Lines"];
			if classesToList > 0 then
				TC_WriteClassList(frame, class, linesLeft);
				writeCount = writeCount - classSize;
				classesWithPlayers = classesWithPlayers - 1;
				classesToList = classesToList - 1;
			elseif (classesWithPlayers > 1) and (linesLeft == 1) then
				TC_WriteMore(frame, writeCount);
				return;
			else
				TC_WriteClassSummary(frame, class);
				writeCount = writeCount - classSize;
				classesWithPlayers = classesWithPlayers - 1;
			end;
		end
	end;
end;

TC_WriteFull = function (frame)
	local unitList = frame["unitList"];
	local writeCondition = not frame["Settings"]["Invert"];
	local writeCountSpecifier;
	if writeCondition then
		writeCountSpecifier = "unitTCount";
	else
		writeCountSpecifier = "unitFCount";
	end;
	local lines = 0;
	local size = unitList[writeCountSpecifier];
	for _, class in ipairs(frame["Settings"]["SortOrder"]) do
		local linesLeft = frame["Settings"]["Limit"]-frame["Lines"];
		local classSize = unitList[class][writeCountSpecifier];
		if (linesLeft < classSize) or ((linesLeft == classSize) and (size - classSize - lines > 1)) then
			lines = lines + TC_WriteClassList(frame, class, linesLeft-1);
			TC_WriteMore(frame, size-lines);
			frame:finishWrites();
			return;
		else
			lines = lines + TC_WriteClassList(frame, class, linesLeft);
		end;
	end;
end;

TC_Write = TC_WriteReduced;

TC_WriteClassList = function (frame, class, limit)
	local unitList = frame["unitList"][class];
	local writeCondition = not frame["Settings"]["Invert"];
	local color = RAID_CLASS_COLORS[string.upper(class)];
	if color == nil then
		-- Pets and other non-class units
		color = { ["r"] = 0.75; ["g"] = 0.75; ["b"] = 0.75; };
	end;
	local count = limit;
	for _, unit in ipairs(unitList) do
		if (count > 0) then
			if unit[2] == writeCondition then
				frame:writeLine((UnitName(unit[1])), color.r,color.g,color.b, nil, unit[1]);
				count = count - 1;
			end;
		else
			return limit;
		end;
	end;
	return limit-count;
end;

TC_WriteClassSummary = function (frame, class)
	local unitList = frame["unitList"][class];
	local color = RAID_CLASS_COLORS[class];
	local writeCount;
	if not frame["Settings"]["Invert"] then
		writeCount = unitList["unitTCount"];
	else
		writeCount = unitList["unitFCount"];
	end;
	if color == nil then color = { ["r"] = 0.5; ["g"] = 0.5; ["b"] = 0.5; }; end;
	if (writeCount == 1) then
		TC_WriteClassList(frame, class, 1);
	elseif (writeCount > 1) then
		frame:writeLine("> "..writeCount.." "..TARGETCHECK_LOCALIZED_CLASS_PLURAL[class],
			color.r,color.g,color.b);
	end;
	return writeCount;
end;

TC_WriteMore = function (frame, num)
	if (num > 0) then
		frame:writeLine("> "..num.." more...", 0.8,0.8,0.8);
	end;
end;

TC_WriteLine = function (frame, text, cr, cg, cb, idx, unit)
	if (not idx) then
		if frame["Lines"] <= 40 then
			idx = frame["Lines"];
		else
			idx = 40;
		end;
	end;
	local textLine = getglobal(frame["Name"].."_Text"..idx);
	local texLine = getglobal(frame["Name"].."_Texture"..idx);
	textLine:SetText(text);
	textLine:SetTextColor(cr, cg, cb);
	textLine:Show();

	if unit and frame["Settings"]["Mana"] then
		local width = (frame["Pointer"]:GetWidth()-16)*UnitMana(unit)/UnitManaMax(unit);
		texLine:ClearAllPoints();
		texLine:SetPoint("TOPLEFT", frame["Pointer"], "TOPLEFT", 8, (idx*-14)-8);
		texLine:SetPoint("BOTTOMRIGHT", frame["Pointer"], "TOPLEFT", 8+width, (idx*-14)-22);
		texLine:Show();
	end;

	frame["Lines"] = frame["Lines"] + 1;
end;

TC_WriteLineTooltip = function(frame, text, cr, cg, cb)
	frame["Pointer"]:AddLine(text, cr, cg, cb, 0.5);
	frame["Lines"] = frame["Lines"] + 1;
end;

TC_FinishWrites = function (frame)
	if not TC_Resizing then
		frame["Pointer"]:SetHeight(frame["Lines"]*14+16);
	end;
	if not TC_Transforming and frame["Lines"] == 0 then
		frame["Pointer"]:Hide();
	else
		frame["Pointer"]:Show();
	end;
end;

TC_FinishWritesTooltip = function (frame)
	if frame["Lines"] == 0 then
		frame["Pointer"]:Hide();
	else
		frame["Pointer"]:Show();
	end;
end;

TC_ClearLines = function (frame)
	frame["Lines"] = 0;
	for i = 0, 40 do
		local text = getglobal(frame["Name"].."_Text"..i);
		local tex = getglobal(frame["Name"].."_Texture"..i);
		text:Hide();
		tex:Hide();
	end;
end;

TC_ClearLinesTooltip = function (frame)
	frame["Lines"] = 0;
	if not (frame["Pointer"]:IsOwned(UIParent)) then
		frame["Pointer"]:SetOwner(UIParent, "ANCHOR_CURSOR");
		frame["Pointer"]:SetAlpha(TC_Settings["Alpha"]*0.5);
		frame["Pointer"]:SetBackdropColor(0,0,0,TC_Settings["Alpha"]*0.5);
		frame["Pointer"]:SetBackdropBorderColor(1,1,1 ,TC_Settings["Alpha"]*0.5);
	end;
	frame["Pointer"]:ClearLines();
end;

-- Moves the given class to the front of the array
TC_ChangeSortOrder = function (array, class)
	local pos = 1;
	while array[pos] ~= class and array[pos] ~= nil do
		pos = pos + 1;
	end;
	if array[pos] == class then -- there should never be an else
		table.remove(array, pos);
		table.insert(array, 1, class);
		TC_MSG("TC: "..TARGETCHECK_LOCALIZED_CLASS[class].." moved to the front.");
		TC_Check();
	end;
end;

TC_TogglePref = function (array, setting, state)
	if state == "off" or state == "false" then
		array[setting] = false;
		TC_MSG("TC: "..setting.." Disabled");
	elseif state == "on" or state == "true" then
		array[setting] = true;
		TC_MSG("TC: "..setting.." Enabled");
	else
		array[setting] = not array[setting];
		if array[setting] then
			TC_MSG("TC: "..setting.." Enabled");
		else
			TC_MSG("TC: "..setting.." Disabled");
		end;
	end;
	TC_Check();
end;

TC_SetPrefNum = function (array, setting, value)
	local valueNum = tonumber(value);
	if valueNum == nil then
		TC_MSG("BT: "..setting.." Input Error, "..value.." is not a number");
	else
		array[setting] = valueNum;
		TC_MSG("BT: "..setting.." set to "..valueNum);
	end;
	TC_Check();
end;

TC_SetAnchor = function (frame, anchor)
	local xp, yp;
	frame["Settings"]["Anchor"] = anchor;
	if string.find(anchor, "TOP") then
		yp = frame["Pointer"]:GetTop();
	else
		yp = frame["Pointer"]:GetBottom();
	end;
	if string.find(anchor, "LEFT") then
		xp = frame["Pointer"]:GetLeft();
	else
		xp = frame["Pointer"]:GetRight();
	end;
	frame["Pointer"]:ClearAllPoints();
	frame["Pointer"]:SetPoint(anchor, UIParent, "BOTTOMLEFT", xp, yp);
	TC_Check();
end;

TC_SetFrameTarget = function (frame, target)
	frame["Settings"]["Target"] = target;
	TC_Check();
end;

TC_SetFrameTargetCustom = function (frame)
	TargetCheck_TextQuery.func = TC_SetFrameTarget;
	TargetCheck_TextQuery.arg1 = frame;
	TargetCheck_TextQuery:Show();
end;

TC_SlashHandler = function (msg)
 	local Cmd, SubCmd = TC_GetCmd(msg);
 	Cmd = string.lower(Cmd);
	if (Cmd == "scale") then
		TC_SetPrefNum(TC_Settings, "Scale", SubCmd);
		for _, f in ipairs(TC_FrameList) do
			f["Pointer"]:SetScale(TC_Settings["Scale"]);
		end
	elseif (Cmd == "alpha") then
		TC_SetPrefNum(TC_Settings, "Alpha", SubCmd);
		for _, f in ipairs(TC_FrameList) do
			f["Pointer"]:SetAlpha(TC_Settings["Alpha"]);
		end
	elseif (Cmd == "limit") then
		for _, f in ipairs(TC_FrameList) do TC_SetPrefNum(f["Settings"], "Limit", SubCmd); end;
	elseif (Cmd == "filter") then
		local class, state = TC_GetCmd(SubCmd);
		class = string.upper(class);
		if (TC_ClassList[class]) then
			for _, f in ipairs(TC_FrameList) do TC_TogglePref(f["Settings"], class, state); end;
		else
			TC_MSG("TC: Class "..class.." can not be filtered out");
		end;
	elseif (Cmd == "sort") then
		local class, state = TC_GetCmd(SubCmd);
		class = string.upper(class);
		if (TC_ClassList[class]) then
			for _, f in ipairs(TC_FrameList) do TC_ChangeSortOrder(f["Settings"]["SortOrder"], class); end;
		else
			TC_MSG("TC: "..class.." is not a valid class.");
		end;
	elseif (Cmd == "header") then
		for _, f in ipairs(TC_FrameList) do TC_TogglePref(f["Settings"], "Header", SubCmd); end;
	elseif (Cmd == "mana") then
		for _, f in ipairs(TC_FrameList) do TC_TogglePref(f["Settings"], "Mana", SubCmd); end;
	elseif (Cmd == "invert") then
		for _, f in ipairs(TC_FrameList) do TC_TogglePref(f["Settings"], "Invert", SubCmd); end;
	elseif (Cmd == "anchor") then
		if (SubCmd == "topleft") or (SubCmd == "topright") or
			(SubCmd == "bottomleft") or (SubCmd == "bottomright") then
			for _, f in ipairs(TC_FrameList) do
				TC_SetAnchor(f, string.upper(SubCmd));
			end;
			TC_MSG("TC: Anchor set to "..SubCmd);
		else
			TC_MSG("TC: Anchor could not be set to "..SubCmd);
		end;
	elseif (Cmd == "mouse") then
		TC_TogglePref(TC_Settings, "Mouse", SubCmd);
		if (TC_Settings["Mouse"]) then
		else
			TC_MouseFrame["Pointer"]:Hide();
		end;
	elseif (Cmd == "mouseheader") then
		TC_TogglePref(TC_Settings["MouseFrame"], "Header", SubCmd);
	elseif (Cmd == "mouselimit") then
		TC_SetPrefNum(TC_Settings["MouseFrame"], "Limit", SubCmd);
	elseif (Cmd == "mousefilter") then
		local class, state = TC_GetCmd(SubCmd);
		class = string.upper(class);
		if (TC_ClassList[class]) then
			TC_TogglePref(TC_Settings["MouseFrame"], class, state);
		else
			TC_MSG("TC: Class "..class.." can not be filtered out");
		end;
	elseif (Cmd == "mousesort") then
		local class, state = TC_GetCmd(SubCmd);
		class = string.upper(class);
		if (TC_ClassList[class]) then
			TC_ChangeSortOrder(TC_Settings["MouseFrame"]["SortOrder"], class);
		else
			TC_MSG("TC: "..class.." is not a valid class.");
		end;
	elseif (Cmd == "create") then
		local f = TC_CreateFrame();
		TC_RestoreFrame(f, table.getn(TC_FrameList)+1);
 	else
 		TC_MSG("|cffccccccTargetCheck v"..TC_Version);
 		TC_MSG("Shift Click to Move, Alt Click to resize");
 		TC_MSG("Right Click to set individual settings");
 		TC_MSG("Use \"targetcheck header on\" to show all frames");
		TC_MSG("/targetcheck scale <num>");
		TC_MSG("/targetcheck alpha <num>");
		TC_MSG("/targetcheck filter <class> [on|off]");
		TC_MSG("/targetcheck sort <class>");
		TC_MSG("/targetcheck header [on|off]");
		TC_MSG("/targetcheck mana [on|off]");
		TC_MSG("/targetcheck invert [on|off]");
		TC_MSG("/targetcheck anchor [topleft|bottomleft]");
		TC_MSG("/targetcheck limit <num>");
		TC_MSG("/targetcheck mouse [on|off]");
		TC_MSG("/targetcheck mouseFilter <class> [on|off]");
		TC_MSG("/targetcheck mouseSort <class>");
		TC_MSG("/targetcheck mouseHeader [on|off]");
		TC_MSG("/targetcheck mouseLimit <num>");
		TC_MSG("/targetcheck create");
 	end;
end;

TC_MSG = function (msg)
	DEFAULT_CHAT_FRAME:AddMessage(msg);
end;

-- Creates lots of temporary arrays. Garbage Collection increase...
TC_Menu_Initialize = function (frame, level)
	if level == 1 then
		UIDropDownMenu_AddButton({
			["text"] = "TargetCheck v"..TC_Version;
			["isTitle"] = 1;
			["notCheckable"] = 1;
			["owner"] = TargetCheck_OutputFrame;
			});
		UIDropDownMenu_AddButton({
			["text"] = "Target";
			["value"] = "Target";
			["notCheckable"] = 1;
			["hasArrow"] = 1;
			});
		UIDropDownMenu_AddButton({
			["text"] = "Filter";
			["value"] = "Filter";
			["notCheckable"] = 1;
			["hasArrow"] = 1;
			});
		UIDropDownMenu_AddButton({
			["text"] = "Sort Order";
			["value"] = "SortOrder";
			["notCheckable"] = 1;
			["hasArrow"] = 1;
			});
		UIDropDownMenu_AddButton({
			["text"] = "Anchor";
			["value"] = "Anchor";
			["notCheckable"] = 1;
			["hasArrow"] = 1;
			});
		UIDropDownMenu_AddButton({
			["text"] = "Show Header";
			["checked"] = TC_CurrentFrameWithMenu["Settings"]["Header"];
			["func"] = TC_Menu_TogglePref;
			["arg1"] = TC_CurrentFrameWithMenu["Settings"];
			["arg2"] = "Header";
			});
		UIDropDownMenu_AddButton({
			["text"] = "Show Mana";
			["checked"] = TC_CurrentFrameWithMenu["Settings"]["Mana"];
			["func"] = TC_Menu_TogglePref;
			["arg1"] = TC_CurrentFrameWithMenu["Settings"];
			["arg2"] = "Mana";
			});
		UIDropDownMenu_AddButton({
			["text"] = "Inverted Display";
			["checked"] = TC_CurrentFrameWithMenu["Settings"]["Invert"];
			["func"] = TC_Menu_TogglePref;
			["arg1"] = TC_CurrentFrameWithMenu["Settings"];
			["arg2"] = "Invert";
			});
		UIDropDownMenu_AddButton({
			["text"] = "Delete";
			["notCheckable"] = 1;
			["func"] = TC_DeleteFrame;
			["arg1"] = TC_CurrentFrameWithMenu;
			});
	elseif level == 2 then
		if UIDROPDOWNMENU_MENU_VALUE == "Filter" then
			for _, class in ipairs(TC_CurrentFrameWithMenu["Settings"]["SortOrder"]) do
				local toggle, checked;
				if TC_CurrentFrameWithMenu["Settings"][class] then
					checked = true;
				else
					checked = false;
				end;
				local color = RAID_CLASS_COLORS[class];
				if color == nil then color = { ["r"] = 1; ["g"] = 1; ["b"] = 1; }; end;
				UIDropDownMenu_AddButton({
					["text"] = TARGETCHECK_LOCALIZED_CLASS[class];
					["checked"] = checked;
					["func"] = TC_Menu_TogglePref;
					["arg1"] = TC_CurrentFrameWithMenu["Settings"];
					["arg2"] = class;
					["keepShownOnClick"] = 1;
				}, 2);
			end;
		elseif UIDROPDOWNMENU_MENU_VALUE == "SortOrder" then
			UIDropDownMenu_AddButton({
				["text"] = "Select to move to the front";
				["notClickable"] = 1;
				["notCheckable"] = 1;
				}, 2);
			for _, class in ipairs(TC_CurrentFrameWithMenu["Settings"]["SortOrder"]) do
				local color = RAID_CLASS_COLORS[class];
				if color == nil then color = { ["r"] = 1; ["g"] = 1; ["b"] = 1; }; end;
				UIDropDownMenu_AddButton({
					["text"] = TARGETCHECK_LOCALIZED_CLASS[class];
					["notCheckable"] = 1;
					["func"] = TC_Menu_ChangeSortOrder;
					["arg1"] = TC_CurrentFrameWithMenu["Settings"]["SortOrder"];
					["arg2"] = class;
				}, 2);
			end;
		elseif UIDROPDOWNMENU_MENU_VALUE == "Target" then
			UIDropDownMenu_AddButton({
				["text"] = "Current Target: "..TC_CurrentFrameWithMenu["Settings"]["Target"];
				["isTitle"] = 1;
				["notCheckable"] = 1;
				}, 2);
			UIDropDownMenu_AddButton({
				["text"] = "player".." ("..(UnitName("player"))..")";
				["notCheckable"] = 1;
				["func"] = TC_Menu_SetFrameTarget;
				["arg1"] = TC_CurrentFrameWithMenu;
				["arg2"] = "player";
				}, 2);
			UIDropDownMenu_AddButton({
				["text"] = "target";
				["notCheckable"] = 1;
				["func"] = TC_Menu_SetFrameTarget;
				["arg1"] = TC_CurrentFrameWithMenu;
				["arg2"] = "target";
				}, 2);
			UIDropDownMenu_AddButton({
				["text"] = "focus";
				["notCheckable"] = 1;
				["func"] = TC_Menu_SetFrameTarget;
				["arg1"] = TC_CurrentFrameWithMenu;
				["arg2"] = "focus";
				}, 2);
			--[[
			UIDropDownMenu_AddButton({
				["text"] = "Custom...";
				["notCheckable"] = 1;
				["func"] = TC_Menu_Exec;
				["arg1"] = TC_SetFrameTargetCustom;
				["arg2"] = TC_CurrentFrameWithMenu;
				}, 2);
			]]
			UIDropDownMenu_AddButton({
				["text"] = "Party";
				["value"] = "Party";
				["notCheckable"] = 1;
				["hasArrow"] = 1;
				}, 2);
			for i = 1,8 do
				UIDropDownMenu_AddButton({
					["text"] = "Raid Party "..i;
					["value"] = i;
					["notCheckable"] = 1;
					["hasArrow"] = 1;
					}, 2);
			end;
		elseif UIDROPDOWNMENU_MENU_VALUE == "Anchor" then
			UIDropDownMenu_AddButton({
				["text"] = "Top Left";
				["notCheckable"] = 1;
				["func"] = TC_Menu_SetAnchor;
				["arg1"] = TC_CurrentFrameWithMenu;
				["arg2"] = "TOPLEFT";
				}, 2);
			UIDropDownMenu_AddButton({
				["text"] = "Bottom Right";
				["notCheckable"] = 1;
				["func"] = TC_Menu_SetAnchor;
				["arg1"] = TC_CurrentFrameWithMenu;
				["arg2"] = "BOTTOMRIGHT";
				}, 2);
		end;
	elseif level == 3 then
		if UIDROPDOWNMENU_MENU_VALUE == "Party" then
			local u,s;
			for i = 1,4 do
				u = "party"..i;
				if UnitExists(u) then
					s = u.." ("..(UnitName(u))..")";
				else
					s = u;
				end;
				UIDropDownMenu_AddButton({
					["text"] = s;
					["notCheckable"] = 1;
					["func"] = TC_Menu_SetFrameTarget;
					["arg1"] = TC_CurrentFrameWithMenu;
					["arg2"] = u;
					}, 3);
			end;
		else
			local j = (UIDROPDOWNMENU_MENU_VALUE-1)*5+1;
			local u, s;
			for i = j,j+4 do
				u = "raid"..i;
				if UnitExists(u) then
					s = u.." ("..(UnitName(u))..")";
				else
					s = u;
				end;
				UIDropDownMenu_AddButton({
					["text"] = s;
					["notCheckable"] = 1;
					["func"] = TC_Menu_SetFrameTarget;
					["arg1"] = TC_CurrentFrameWithMenu;
					["arg2"] = u;
					}, 3);
			end;
		end;
	end
end;

TC_Menu_Exec = function (self, func, arg)
	func(arg);
end;

TC_Menu_TogglePref = function (self, setting, value)
	TC_TogglePref(setting, value);
end;

TC_Menu_SetAnchor = function (self, frame, anchor)
	TC_SetAnchor(frame, anchor);
end;

TC_Menu_ChangeSortOrder = function (self, frame, class)
	TC_ChangeSortOrder(frame, class);
end;

TC_Menu_SetFrameTarget = function (self, frame, target)
	TC_SetFrameTarget(frame, target);
end;

--[[
List of button attributes
======================================================
info.text = [STRING]  --  The text of the button
info.value = [ANYTHING]  --  The value that UIDROPDOWNMENU_MENU_VALUE is set to when the button is clicked
info.func = [function()]  --  The function that is called when you click the button
info.checked = [nil, true, function]  --  Check the button if true or function returns true
info.isTitle = [nil, true]  --  If it's a title the button is disabled and the font color is set to yellow
info.disabled = [nil, true]  --  Disable the button and show an invisible button that still traps the mouseover event so menu doesn't time out
info.hasArrow = [nil, true]  --  Show the expand arrow for multilevel menus
info.hasColorSwatch = [nil, true]  --  Show color swatch or not, for color selection
info.r = [1 - 255]  --  Red color value of the color swatch
info.g = [1 - 255]  --  Green color value of the color swatch
info.b = [1 - 255]  --  Blue color value of the color swatch
info.colorCode = [STRING] -- "|cAARRGGBB" embedded hex value of the button text color. Only used when button is enabled
info.swatchFunc = [function()]  --  Function called by the color picker on color change
info.hasOpacity = [nil, 1]  --  Show the opacity slider on the colorpicker frame
info.opacity = [0.0 - 1.0]  --  Percentatge of the opacity, 1.0 is fully shown, 0 is transparent
info.opacityFunc = [function()]  --  Function called by the opacity slider when you change its value
info.cancelFunc = [function(previousValues)] -- Function called by the colorpicker when you click the cancel button (it takes the previous values as its argument)
info.notClickable = [nil, 1]  --  Disable the button and color the font white
info.notCheckable = [nil, 1]  --  Shrink the size of the buttons and don't display a check box
info.owner = [Frame]  --  Dropdown frame that "owns" the current dropdownlist
info.keepShownOnClick = [nil, 1]  --  Don't hide the dropdownlist after a button is clicked
info.tooltipTitle = [nil, STRING] -- Title of the tooltip shown on mouseover
info.tooltipText = [nil, STRING] -- Text of the tooltip shown on mouseover
info.justifyH = [nil, "CENTER"] -- Justify button text
info.arg1 = [ANYTHING] -- This is the first argument used by info.func
info.arg2 = [ANYTHING] -- This is the second argument used by info.func
info.fontObject = [FONT] -- font object replacement for Normal and Highlight
info.menuTable = [TABLE] -- This contains an array of info tables to be displayed as a child menu
]]


TC_Menu_Load = function (menu)
	UIDropDownMenu_Initialize(menu, TC_Menu_Initialize, "MENU");
end;

TC_Transforming = false;
TC_Moving = false;
TC_Resizing = false;
TC_FrameMouseDown = function (self, button)
	if (button == "LeftButton") then
		if (IsShiftKeyDown()) then
			self:StartMoving();
			TC_Transforming = true;
			TC_Moving = true;
		elseif (IsAltKeyDown()) then
			local w = self:GetWidth();
			local h = self:GetHeight();
			local x, y = self:GetCenter();
			local mx, my = GetCursorPosition();
			local scaleL, scaleW;
			scaleL = self:GetScale();
			scaleW = UIParent:GetScale();
			mx = mx/scaleW;
			my = my/scaleW;
			x = x*scaleL;
			y = y*scaleL
			x = mx - x;
			y = my - y;
			x = x/w*h;
			if (abs(x) > abs(y)) then
				if (x > 0) then
					self:StartSizing("RIGHT");
				else
					self:StartSizing("LEFT");
				end;
			else
				if (y > 0) then
					self:StartSizing("TOP");
				else
					self:StartSizing("BOTTOM");
				end;
			end;
			TC_Transforming = true;
			TC_Resizing = true;
		end;
	elseif (button == "RightButton") then
		local menu = getglobal(self["TC_Frame"]["Name"].."_Menu");
		TC_CurrentFrameWithMenu = self["TC_Frame"];
		ToggleDropDownMenu(1, nil, menu, "cursor");
	end;
end;

TC_FrameResized = function (frame)
	local height = frame:GetHeight();
	if TC_Resizing then
		TC_SetPrefNum(frame["TC_Frame"]["Settings"], "Limit", math.ceil((height-16)/14));
	end;
	TC_Transforming = false;
	TC_Moving = false;
	TC_Resizing = false;
end;

-- Some CMD parsing code, written by Tigerheart on http://www.wowwiki.com/HOWTO:_Extract_Info_from_a_Slash_Command
TC_GetCmd = function (msg)
	if msg then
		local a,b,c=strfind(msg, "(%S+)"); --contiguous string of non-space characters
		if a then
			return c, strsub(msg, b+2);
		else
			return "";
		end
	end
end;

TC_GetArgument = function (msg)
	if msg then
		local a,b=strfind(msg, "=");
		if a then
			return strsub(msg,1,a-1), strsub(msg, b+1);
		else
			return "";
		end
	end
end;

TC_OnLoad();
